(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('standardized-audio-context')) :
  typeof define === 'function' && define.amd ? define(['exports', 'standardized-audio-context'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.Phonograph = {}, global.standardizedAudioContext));
})(this, (function (exports, standardizedAudioContext) { 'use strict';

  function slice(view, start, end) {
    if (view.slice) {
      return view.slice(start, end);
    }

    const clone = new Uint8Array(end - start);
    let p = 0;

    for (let i = start; i < end; i += 1) {
      // eslint-disable-next-line no-plusplus
      clone[p++] = view[i];
    }

    return clone;
  }

  function _nullishCoalesce$5(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } } function _optionalChain$1(ops) { let lastAccessLHS = undefined; let value = ops[0]; let i = 1; while (i < ops.length) { const op = ops[i]; const fn = ops[i + 1]; i += 2; if ((op === 'optionalAccess' || op === 'optionalCall') && value == null) { return undefined; } if (op === 'access' || op === 'optionalAccess') { lastAccessLHS = value; value = fn(value); } else if (op === 'call' || op === 'optionalCall') { value = fn((...args) => value.call(lastAccessLHS, ...args)); lastAccessLHS = undefined; } } return value; }




  class Chunk {
    

    

    __init() {this.duration = null;}

    __init2() {this.numFrames = null;}

    

    

    

    

    

    __init3() {this.next = null;}

    

    

    

     __init4() {this.callbacks = {};}

    constructor({
      clip,
      raw,
      onready,
      onerror,
      adapter,
      chunkIndex,
    }






  ) {Chunk.prototype.__init.call(this);Chunk.prototype.__init2.call(this);Chunk.prototype.__init3.call(this);Chunk.prototype.__init4.call(this);
      this.clip = clip;
      this.chunkIndex = chunkIndex;
      this.context = clip.context;

      this.raw = raw;
      this.extended = null;
      this.extendedWithHeader = null;

      this.adapter = adapter;

      this.duration = null;
      this.ready = false;

      this._attached = false;
      if (onready !== null) {
        this.once('ready', onready);
      }
      this.once('error', onerror);

      this._firstByte = 0;

      const decode = (callback, errback) => {
        const buffer = slice(raw, this._firstByte, raw.length);
        const bufferWithId3Header = this.adapter.wrapChunk(buffer).buffer;

        this.context.decodeAudioData(bufferWithId3Header, callback, (err) => {
          if (err) {
            return errback(err);
          }

          this._firstByte += 1;

          // filthy hack taken from http://stackoverflow.com/questions/10365335/decodeaudiodata-returning-a-null-error
          // Thanks Safari developers, you absolute numpties
          for (; this._firstByte < raw.length - 1; this._firstByte += 1) {
            if (this.adapter.validateChunk(raw, this._firstByte)) {
              return decode(callback, errback);
            }
          }

          errback(new Error(`Could not decode audio buffer`));
        });
      };

      decode(
        () => {
          let numFrames = 0;
          let duration = 0;
          let i = this._firstByte;

          while (i < this.raw.length) {
            if (this.adapter.validateChunk(this.raw, i)) {
              const metadata = this.adapter.getChunkMetadata(this.raw, i);
              numFrames += 1;

              const frameLength = this.adapter.getChunkLength(
                this.raw,
                metadata,
                i
              );
              i += Math.max(_nullishCoalesce$5(frameLength, () => ( 0)), 4);

              // @ts-ignore
              (window ).sampleRates.push(_optionalChain$1([metadata, 'optionalAccess', _ => _.sampleRate]));
              duration +=
                _nullishCoalesce$5(this.adapter.getChunkDuration(this.raw, metadata, i), () => ( 0));
            } else {
              i += 1;
            }
          }

          this.duration = duration;
          this.numFrames = numFrames;

          console.log(this.duration, this.numFrames);
          this._ready();
        },
        () => {
          this._fire('error', new Error(`Could not decode audio buffer`));
        }
      );
    }

    off(eventName, cb) {
      const callbacks = this.callbacks[eventName];
      if (!callbacks) return;

      const index = callbacks.indexOf(cb);
      if (~index) callbacks.splice(index, 1);
    }

    on(eventName, cb) {
      const callbacks =
        this.callbacks[eventName] || (this.callbacks[eventName] = []);
      callbacks.push(cb);

      return {
        cancel: () => this.off(eventName, cb),
      };
    }

    once(eventName, cb) {
      const _cb = (data) => {
        cb(data);
        this.off(eventName, _cb);
      };

      return this.on(eventName, _cb);
    }

     _fire(eventName, data) {
      const callbacks = this.callbacks[eventName];
      if (!callbacks) return;

      callbacks.slice().forEach((cb) => cb(data));
    }

    attach(nextChunk) {
      this.next = nextChunk;
      this._attached = true;

      this._ready();
    }

    createBufferCallback(
      callback,
      errback
    ) {
      this.createBuffer().then(callback, errback);
    }

    createBuffer() {
      if (!this.ready) {
        throw new Error(
          'Something went wrong! Chunk was not ready in time for playback'
        );
      }
      return this.context.decodeAudioData(
        slice(this.extendedWithHeader, 0, this.extendedWithHeader.length).buffer
      );
    }

     _ready() {
      if (this.ready) return;

      if (this._attached && this.duration !== null) {
        this.ready = true;

        if (this.next) {
          const rawLen = this.raw.length;
          const nextLen = this.next.raw.length >> 1; // we don't need the whole thing

          this.extended = new Uint8Array(rawLen + nextLen);

          let p = 0;

          for (let i = this._firstByte; i < rawLen; i += 1) {
            // eslint-disable-next-line no-plusplus
            this.extended[p++] = this.raw[i];
          }

          for (let i = 0; i < nextLen; i += 1) {
            // eslint-disable-next-line no-plusplus
            this.extended[p++] = this.next.raw[i];
          }
        } else {
          this.extended =
            this._firstByte > 0
              ? slice(this.raw, this._firstByte, this.raw.length)
              : this.raw;
        }
        this.extendedWithHeader = this.adapter.wrapChunk(this.extended);

        console.log(this.extendedWithHeader);

        this._fire('ready');
      }
    }
  }

  function _nullishCoalesce$4(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } }function warn(msg) {
    console.warn(
      // eslint-disable-line no-console
      `%c🔊 Phonograph.js %c${msg}`,
      'font-weight: bold;',
      'font-weight: normal;'
    );

    console.groupCollapsed(
      // eslint-disable-line no-console
      `%c🔊 stack trace`,
      'font-weight: normal; color: #666;'
    );

    const stack = (_nullishCoalesce$4(new Error().stack, () => ( ''))).split('\n').slice(2).join('\n');

    console.log(
      // eslint-disable-line no-console
      `%c${stack}`,
      'display: block; font-weight: normal; color: #666;'
    );

    console.groupEnd(); // eslint-disable-line no-console
  }

  function _nullishCoalesce$3(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } }










  class FetchLoader  {
    

    

    constructor(url) {
      this.url = url;
      this._cancelled = false;
    }

    cancel() {
      this._cancelled = true;
    }

    load({
      onprogress,
      ondata,
      onload,
      onerror,
    }




  ) {
      this._cancelled = false;

      fetch(this.url)
        .then((response) => {
          if (this._cancelled) return;

          if (!response.ok) {
            onerror(
              new Error(
                `Bad response (${response.status} – ${response.statusText})`
              )
            );
            return;
          }

          const total = +(_nullishCoalesce$3(response.headers.get('content-length'), () => ( 0))) || 0;

          let length = 0;
          onprogress((total ? length : 0) / total, length, total);

          if (response.body) {
            const reader = response.body.getReader();

            const read = () => {
              if (this._cancelled) return;

              reader
                .read()
                .then((chunk) => {
                  if (this._cancelled) return;

                  if (chunk.done) {
                    onprogress(1, length, length);
                    onload();
                  } else {
                    length += chunk.value.length;
                    ondata(chunk.value);
                    onprogress((total ? length : 0) / total, length, total);

                    read();
                  }
                })
                .catch(onerror);
            };

            read();
          } else {
            // Firefox doesn't yet implement streaming
            response
              .arrayBuffer()
              .then((arrayBuffer) => {
                if (this._cancelled) return;

                const uint8Array = new Uint8Array(arrayBuffer);

                ondata(uint8Array);
                onprogress(1, uint8Array.length, uint8Array.length);
                onload();
              })
              .catch(onerror);
          }
        })
        .catch(onerror);
    }
  }

  class XhrLoader  {
    

    

    

    constructor(url) {
      this.url = url;

      this._cancelled = false;
      this._xhr = null;
    }

    cancel() {
      if (this._cancelled) return;

      this._cancelled = true;

      if (this._xhr) {
        this._xhr.abort();
        this._xhr = null;
      }
    }

    load({
      onprogress,
      ondata,
      onload,
      onerror,
    }




  ) {
      this._cancelled = false;

      const xhr = new XMLHttpRequest();
      xhr.responseType = 'arraybuffer';

      xhr.onerror = onerror;

      xhr.onload = (e) => {
        if (this._cancelled) return;

        onprogress(e.loaded / e.total, e.loaded, e.total);
        ondata(new Uint8Array(xhr.response));
        onload();

        this._xhr = null;
      };

      xhr.onprogress = (e) => {
        if (this._cancelled) return;

        onprogress(e.loaded / e.total, e.loaded, e.total);
      };

      xhr.open('GET', this.url);
      xhr.send();

      this._xhr = xhr;
    }
  }

  function _nullishCoalesce$2(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } } function _optionalChain(ops) { let lastAccessLHS = undefined; let value = ops[0]; let i = 1; while (i < ops.length) { const op = ops[i]; const fn = ops[i + 1]; i += 2; if ((op === 'optionalAccess' || op === 'optionalCall') && value == null) { return undefined; } if (op === 'access' || op === 'optionalAccess') { lastAccessLHS = value; value = fn(value); } else if (op === 'call' || op === 'optionalCall') { value = fn((...args) => value.call(lastAccessLHS, ...args)); lastAccessLHS = undefined; } } return value; }/* eslint-disable no-restricted-syntax */

  const CHUNK_SIZE = 64 * 1024;
  const OVERLAP = 0.2;

  class PhonographError extends Error {
    

    

    constructor(message, opts) {
      super(message);

      this.phonographCode = opts.phonographCode;
      this.url = opts.url;
    }
  }

  // A cache of audio buffers starting from current time








  class Clip {
    

    

    

    // eslint-disable-next-line @typescript-eslint/no-explicit-any
     __init() {this.callbacks = {};}

    

    __init2() {this.buffered = 0;}

    __init3() {this.length = 0;}

     __init4() {this._loaded = false;}

     get loaded() {
      return this._loaded;
    }

     set loaded(value) {
      this._loaded = value;
    }

     __init5() {this._canplaythrough = false;}

     get canplaythrough() {
      return this._canplaythrough;
    }

     set canplaythrough(value) {
      this._canplaythrough = value;
    }

    

    __init6() {this.metadata = null;}

    __init7() {this.playing = false;}

    __init8() {this.ended = false;}

     __init9() {this._startTime = 0;}

     __init10() {this._currentTime = 0;}

     __init11() {this.__chunks = [];}

     get _chunks() {
      return this.__chunks;
    }

     set _chunks(value) {
      this.__chunks = value;
    }

     __init12() {this._contextTimeAtStart = 0;}

     __init13() {this._pendingSourceStart = 0;}

     __init14() {this._connected = false;}

    

    

     __init15() {this._loadStarted = false;}

     __init16() {this._actualPlaying = false;}

     get stuck() {
      if (this.playing) {
        return !this._actualPlaying;
      }
      return !this.audioBufferCacheHit();
    }

     __init17() {this._currentSource = null;}

     __init18() {this._nextSource = null;}

     __init19() {this._currentGain = null;}

     __init20() {this._nextGain = null;}

     __init21() {this._audioBufferCache = null;}

    constructor({
      context,
      url,
      loop,
      volume,
      adapter,
    }





  ) {Clip.prototype.__init.call(this);Clip.prototype.__init2.call(this);Clip.prototype.__init3.call(this);Clip.prototype.__init4.call(this);Clip.prototype.__init5.call(this);Clip.prototype.__init6.call(this);Clip.prototype.__init7.call(this);Clip.prototype.__init8.call(this);Clip.prototype.__init9.call(this);Clip.prototype.__init10.call(this);Clip.prototype.__init11.call(this);Clip.prototype.__init12.call(this);Clip.prototype.__init13.call(this);Clip.prototype.__init14.call(this);Clip.prototype.__init15.call(this);Clip.prototype.__init16.call(this);Clip.prototype.__init17.call(this);Clip.prototype.__init18.call(this);Clip.prototype.__init19.call(this);Clip.prototype.__init20.call(this);Clip.prototype.__init21.call(this);Clip.prototype.__init22.call(this);
      this.context = context || new standardizedAudioContext.AudioContext();
      this.url = url;
      this.loop = loop || false;
      this.adapter = adapter;

      this.loader = new ((window ).fetch ? FetchLoader : XhrLoader)(url);

      this.fadeTarget = volume || 1;
      this._gain = this.context.createGain();
      this._gain.gain.value = this.fadeTarget;

      this._gain.connect(this.context.destination);

      this._chunks = [];
    }

    buffer(bufferToCompletion = false) {
      if (!this._loadStarted) {
        this._loadStarted = true;

        const tempBuffer = new Uint8Array(CHUNK_SIZE * 2);
        let p = 0;
        let processedBytes = 0;
        let nextFrameStartBytes = 0;

        const loadStartTime = Date.now();
        let totalLoadedBytes = 0;

        const checkCanplaythrough = () => {
          if (this.canplaythrough || !this.length) return;

          let duration = 0;
          let bytes = 0;

          for (const chunk of this._chunks) {
            if (!chunk.duration) break;
            duration += chunk.duration;
            bytes += chunk.raw.length;
          }

          if (!duration) return;

          const scale = this.length / bytes;
          const estimatedDuration = duration * scale;

          const timeNow = Date.now();
          const elapsed = timeNow - loadStartTime;

          const bitrate = totalLoadedBytes / elapsed;
          const estimatedTimeToDownload =
            (1.5 * (this.length - totalLoadedBytes)) / bitrate / 1e3;

          // if we have enough audio that we can start playing now
          // and finish downloading before we run out, we've
          // reached canplaythrough
          const availableAudio = (bytes / this.length) * estimatedDuration;

          if (availableAudio > estimatedTimeToDownload) {
            this.canplaythrough = true;
            this._fire('canplaythrough');
          }
        };

        const drainBuffer = () => {
          const firstByte = 0;

          const chunk = new Chunk({
            clip: this,
            chunkIndex: this.__chunks.length,
            raw: slice(tempBuffer, firstByte, p),
            onready: () => {
              if (!this.canplaythrough) {
                checkCanplaythrough();
              }
              this.trySetupAudioBufferCache();
            },
            onerror: (error) => {
              error.url = this.url;
              error.phonographCode = 'COULD_NOT_DECODE';
              this._fire('loaderror', error);
            },
            adapter: this.adapter,
          });

          const lastChunk = this._chunks[this._chunks.length - 1];
          if (lastChunk) lastChunk.attach(chunk);

          this._chunks.push(chunk);
          processedBytes += p;
          p = 0;

          return chunk;
        };

        this.loader.load({
          onprogress: (progress, length, total) => {
            this.buffered = length;
            this.length = total;
            this._fire('loadprogress', { progress, length, total });
          },

          ondata: (uint8Array) => {
            if (!this.metadata) {
              for (let i = 0; i < uint8Array.length; i += 1) {
                // determine some facts about this mp3 file from the initial header
                if (this.adapter.validateChunk(uint8Array)) {
                  const metadata = this.adapter.getChunkMetadata(uint8Array);
                  this.metadata = metadata;

                  break;
                }
              }
            }

            for (let i = 0; i < uint8Array.length; i += 1) {
              // once the buffer is large enough, wait for
              // the next frame header then drain it
              // TODO: header detection may not works if the header is divided into two data segments.
              if (
                p + processedBytes >= nextFrameStartBytes &&
                this.adapter.validateChunk(uint8Array, i)
              ) {
                const metadata = this.adapter.getChunkMetadata(uint8Array, i);
                nextFrameStartBytes =
                  p +
                  processedBytes +
                  (_nullishCoalesce$2(this.adapter.getChunkLength(uint8Array, metadata, i), () => ( 0)));
                if (p > CHUNK_SIZE + 4) {
                  drainBuffer();
                }
              }

              // write new data to buffer
              tempBuffer[p] = uint8Array[i];
              p += 1;
            }

            totalLoadedBytes += uint8Array.length;
          },

          onload: () => {
            if (p) {
              const lastChunk = drainBuffer();
              lastChunk.attach(null);

              totalLoadedBytes += p;
            }

            const actualLoad = () => {
              if (!this.canplaythrough) {
                this.canplaythrough = true;
                this._fire('canplaythrough');
              }

              this.loaded = true;
              this._fire('load');
            };
            if (this._chunks[0].ready) {
              actualLoad();
            } else {
              this._chunks[0].once('ready', actualLoad);
            }
          },

          onerror: (error) => {
            error.url = this.url;
            error.phonographCode = 'COULD_NOT_LOAD';
            this._fire('loaderror', error);
            this._loadStarted = false;
          },
        });
      }

      return new Promise((fulfil, reject) => {
        const ready = bufferToCompletion ? this.loaded : this.canplaythrough;

        if (ready) {
          fulfil();
        } else {
          this.once(bufferToCompletion ? 'load' : 'canplaythrough', () => {
            fulfil();
          });
          this.once('loaderror', reject);
          this.once('dispose', reject);
        }
      });
    }

    clone() {
      return new Clone(this);
    }

    connect(
      destination,
      output,
      input
    ) {
      if (!this._connected) {
        this._gain.disconnect();
        this._connected = true;
      }

      this._gain.connect(destination, output, input);
      return this;
    }

    disconnect(
      destination,
      output,
      input
    ) {
      this._gain.disconnect(destination, output, input);
    }

    _disconnectAllAndReplaceAudioContext(newAudioContext) {
      // To be safe, make sure the clip is paused when calling this
      // TODO: do we need to maintain more thing
      this._gain.disconnect();
      this.context = newAudioContext;
      this._chunks.forEach((chunk) => {
        chunk.context = newAudioContext;
      });
      this._gain = newAudioContext.createGain();
      this._gain.gain.value = this.fadeTarget;
    }

    dispose() {
      if (this.playing) this.pause();

      if (this._loadStarted) {
        this.loader.cancel();
        this._loadStarted = false;
      }

      this._currentTime = 0;
      this.loaded = false;
      this.canplaythrough = false;
      this._chunks = [];

      this._fire('dispose');
    }

    off(eventName, cb) {
      const callbacks = this.callbacks[eventName];
      if (!callbacks) return;

      const index = callbacks.indexOf(cb);
      if (~index) callbacks.splice(index, 1);
    }

    on(eventName, cb) {
      const callbacks =
        this.callbacks[eventName] || (this.callbacks[eventName] = []);
      callbacks.push(cb);

      return {
        cancel: () => this.off(eventName, cb),
      };
    }

    once(eventName, cb) {
      const _cb = (data) => {
        cb(data);
        this.off(eventName, _cb);
      };

      return this.on(eventName, _cb);
    }

    play() {
      const promise = new Promise((fulfil, reject) => {
        this.once('ended', fulfil);

        this.once('loaderror', reject);
        this.once('playbackerror', reject);

        this.once('dispose', () => {
          if (this.ended) return;

          const err = new PhonographError('Clip was disposed', {
            phonographCode: 'CLIP_WAS_DISPOSED',
            url: this.url,
          });
          reject(err);
        });
      });

      if (this.playing) {
        warn(
          `clip.play() was called on a clip that was already playing (${this.url})`
        );
      } else if (!this.canplaythrough) {
        warn(
          `clip.play() was called before clip.canplaythrough === true (${this.url})`
        );
        this.buffer();
      }
      this.playing = true;
      this.ended = false;
      this.tryResumePlayback();

      return promise;
    }

    pause() {
      if (!this.playing) {
        warn(
          `clip.pause() was called on a clip that was already paused (${this.url})`
        );
        return this;
      }

      this.resetAudioNodes();
      this.stopFade();
      this.playing = false;
      this._actualPlaying = false;
      this._fire('pause');

      return this;
    }

    get currentTime() {
      if (this.playing && this._actualPlaying) {
        return (
          this._startTime + (this.context.currentTime - this._contextTimeAtStart)
        );
      }
      return this._currentTime;
    }

    set currentTime(currentTime) {
      if (this.playing) {
        this.pause();
        this._currentTime = currentTime;
        this._audioBufferCache = null;
        this.trySetupAudioBufferCache();
        this.play().catch(() => {});
      } else {
        this._currentTime = currentTime;
        this._audioBufferCache = null;
        this.trySetupAudioBufferCache();
      }
    }

    get duration() {
      let total = 0;
      for (const chunk of this._chunks) {
        if (!chunk.duration) return null;
        total += chunk.duration;
      }

      return total;
    }

    get paused() {
      return !this.playing;
    }

    get volume() {
      return this._gain.gain.value;
    }

    set volume(volume) {
      this.stopFade();
      this.fadeTarget = volume;
      this._gain.gain.value = volume;
    }

    fade(startVolume, endVolume, duration) {
      this.stopFade();
      if (!this.playing) {
        this.volume = endVolume;
        return;
      }
      const now = this.context.currentTime;
      this._gain.gain.value = startVolume;
      this._gain.gain.linearRampToValueAtTime(endVolume, now + duration);
      this.fadeTarget = endVolume;
    }

     stopFade() {
      const now = this.context.currentTime;
      this._gain.gain.cancelScheduledValues(now);
      this._gain.gain.value = this.fadeTarget;
    }

     _fire(eventName, data) {
      const callbacks = this.callbacks[eventName];
      if (!callbacks) return;

      callbacks.slice().forEach((cb) => cb(data));
    }

    // Attempt to setup AudioBufferCache if it is not setup
    // Should be called when new chunk is ready
     trySetupAudioBufferCache() {
      if (this._audioBufferCache !== null) {
        return;
      }
      let lastChunk = null;
      let chunk = _nullishCoalesce$2(this._chunks[0], () => ( null));
      let time = 0;
      while (chunk !== null) {
        if (chunk.duration === null) {
          return;
        }
        const chunkEnd = time + chunk.duration;
        if (chunkEnd > this._currentTime) {
          // TODO: reuse audio buffer in old cache
          this._audioBufferCache = {
            currentChunkStartTime: time,
            currentChunk: chunk,
            currentBuffer: null,
            nextBuffer: null,
            pendingBuffer: null,
          };
          this.decodeChunk(_nullishCoalesce$2(chunk, () => ( null)));
          this.decodeChunk(_nullishCoalesce$2(_optionalChain([chunk, 'optionalAccess', _ => _.next]), () => ( null)));
          this.decodeChunk(_nullishCoalesce$2(_optionalChain([chunk, 'optionalAccess', _2 => _2.next, 'optionalAccess', _3 => _3.next]), () => ( null)));
          return;
        }
        time = chunkEnd;
        lastChunk = chunk;
        chunk = lastChunk.next;
      }
      // All available Chunk visited, check if there are more chunks to be load.
      if (_optionalChain([lastChunk, 'optionalAccess', _4 => _4.ready])) {
        this._audioBufferCache = {
          currentChunkStartTime: time,
          currentChunk: null,
          currentBuffer: null,
          nextBuffer: null,
          pendingBuffer: null,
        };
      }
    }

    // Check is there enough audio buffer to schedule current and next chunk
     audioBufferCacheHit() {
      const audioBufferCache = this._audioBufferCache;
      if (audioBufferCache === null) {
        return false;
      }
      const { currentChunk, currentBuffer, nextBuffer } = audioBufferCache;
      if (currentChunk === null) {
        return true;
      }
      if (!currentChunk.ready) {
        return false;
      }
      if (currentBuffer === null) {
        return false;
      }
      const nextChunk = currentChunk.next;
      if (nextChunk === null) {
        return true;
      }
      if (!nextChunk.ready) {
        return false;
      }
      if (nextBuffer === null) {
        return false;
      }
      return true;
    }

    // Advance the audioBufferCache to next chunk when current chunk is played
     advanceAudioBufferCache() {
      const { currentChunk, currentChunkStartTime, nextBuffer, pendingBuffer } =
        this._audioBufferCache;
      this._audioBufferCache = {
        currentChunkStartTime: currentChunkStartTime + currentChunk.duration,
        currentChunk: currentChunk.next,
        currentBuffer: nextBuffer,
        nextBuffer: pendingBuffer,
        pendingBuffer: null,
      };
      this.decodeChunk(_nullishCoalesce$2(_optionalChain([currentChunk, 'optionalAccess', _5 => _5.next, 'optionalAccess', _6 => _6.next, 'optionalAccess', _7 => _7.next]), () => ( null)));
    }

    // Start play the audioBuffer when the audioBufferCacheHit is true
     startPlay() {
      this._startTime = this._currentTime;
      this._actualPlaying = true;
      const { currentChunkStartTime, currentChunk, currentBuffer, nextBuffer } =
        this._audioBufferCache;

      this._contextTimeAtStart = this.context.currentTime;
      if (currentChunk !== null) {
        this._pendingSourceStart =
          this._contextTimeAtStart +
          (currentChunk.duration - (this._startTime - currentChunkStartTime));

        this._currentSource = this.context.createBufferSource();
        this._currentSource.buffer = currentBuffer;
        this._currentGain = this.context.createGain();
        this._currentGain.connect(this._gain);
        this._currentGain.gain.setValueAtTime(
          0,
          this._pendingSourceStart + OVERLAP
        );
        this._currentSource.connect(this._currentGain);
        this._currentSource.start(
          this._contextTimeAtStart,
          this._startTime - currentChunkStartTime
        );
        this._currentSource.stop(this._pendingSourceStart + OVERLAP * 2);
        this._currentSource.addEventListener('ended', this.onCurrentSourceEnd);
        if (currentChunk.next !== null) {
          const pendingStart =
            this._pendingSourceStart + currentChunk.next.duration;
          this._nextSource = this.context.createBufferSource();
          this._nextSource.buffer = nextBuffer;
          this._nextGain = this.context.createGain();
          this._nextGain.connect(this._gain);
          this._nextGain.gain.setValueAtTime(0, this._pendingSourceStart);
          this._nextGain.gain.setValueAtTime(
            1,
            this._pendingSourceStart + OVERLAP
          );
          this._nextSource.connect(this._nextGain);
          this._nextSource.start(this._pendingSourceStart);
          this._nextGain.gain.setValueAtTime(0, pendingStart + OVERLAP);
          this._nextSource.stop(pendingStart + OVERLAP * 2);
          this._pendingSourceStart = pendingStart;
        }
      } else {
        this.pause().currentTime = 0;
        // TODO: schedule playing of first chunk instead of do this
        if (this.loop) {
          this.play();
        } else {
          this.ended = true;
          this._fire('ended');
        }
      }
    }

    // Advance audio nodes to next chunk when current chunk is played
    // and the audioBufferCacheHit is true
    // should be called after advanceAudioBufferCache
     advanceAudioNodes() {
      _optionalChain([this, 'access', _8 => _8._currentSource, 'optionalAccess', _9 => _9.stop, 'call', _10 => _10()]);
      _optionalChain([this, 'access', _11 => _11._currentSource, 'optionalAccess', _12 => _12.disconnect, 'call', _13 => _13()]);
      _optionalChain([this, 'access', _14 => _14._currentGain, 'optionalAccess', _15 => _15.disconnect, 'call', _16 => _16()]);
      this._currentGain = this._nextGain;
      this._currentSource = this._nextSource;
      _optionalChain([this, 'access', _17 => _17._currentSource, 'optionalAccess', _18 => _18.addEventListener, 'call', _19 => _19('ended', this.onCurrentSourceEnd)]);
      const { currentChunk, nextBuffer } = this._audioBufferCache;
      if ((_nullishCoalesce$2(_optionalChain([currentChunk, 'optionalAccess', _20 => _20.next]), () => ( null))) !== null) {
        const pendingStart =
          this._pendingSourceStart + currentChunk.next.duration;
        this._nextSource = this.context.createBufferSource();
        this._nextSource.buffer = nextBuffer;
        this._nextGain = this.context.createGain();
        this._nextGain.connect(this._gain);
        this._nextGain.gain.setValueAtTime(0, this._pendingSourceStart);
        this._nextGain.gain.setValueAtTime(1, this._pendingSourceStart + OVERLAP);
        this._nextSource.connect(this._nextGain);
        this._nextSource.start(this._pendingSourceStart);
        this._nextGain.gain.setValueAtTime(0, pendingStart + OVERLAP);
        this._nextSource.stop(pendingStart + OVERLAP * 2);
        this._pendingSourceStart = pendingStart;
      } else {
        this._nextGain = null;
        this._nextSource = null;
      }
      if (currentChunk === null) {
        this.pause().currentTime = 0;
        // TODO: schedule playing of first chunk instead of do this
        if (this.loop) {
          this.play();
        } else {
          this.ended = true;
          this._fire('ended');
        }
      }
    }

    // Reset audioNodes when stuck or paused
     resetAudioNodes() {
      if (this._currentSource) {
        this._currentSource.stop();
        this._currentSource.disconnect();
        this._currentSource = null;
      }
      if (this._nextSource) {
        this._nextSource.stop();
        this._nextSource.disconnect();
        this._nextSource = null;
      }
      if (this._currentGain) {
        this._currentGain.disconnect();
        this._currentGain = null;
      }
      if (this._nextGain) {
        this._nextGain.disconnect();
        this._nextSource = null;
      }
      if (!this.playing || !this._actualPlaying) {
        return;
      }
      this._currentTime =
        this._startTime + (this.context.currentTime - this._contextTimeAtStart);
      this._audioBufferCache = null;
      this.trySetupAudioBufferCache();
    }

    // Advance to next Chunk if playback of current source ends
     __init22() {this.onCurrentSourceEnd = (event) => {
      if (event.target !== this._currentSource) {
        return;
      }
      if (!this.playing || !this._actualPlaying) {
        return;
      }
      this.advanceAudioBufferCache();
      if (this.audioBufferCacheHit()) {
        this.advanceAudioNodes();
      } else {
        this.resetAudioNodes();
        this._actualPlaying = false;
      }
    };}

    // Put audio buffer into AudioBufferCache when it is decoded
     onBufferDecoded(chunk, buffer) {
      const audioBufferCache = this._audioBufferCache;
      if (audioBufferCache === null) {
        return;
      }
      if (audioBufferCache.currentChunk === chunk) {
        if (audioBufferCache.currentBuffer === null) {
          audioBufferCache.currentBuffer = buffer;
        }
      }
      if (_optionalChain([audioBufferCache, 'access', _21 => _21.currentChunk, 'optionalAccess', _22 => _22.next]) === chunk) {
        if (audioBufferCache.nextBuffer === null) {
          audioBufferCache.nextBuffer = buffer;
        }
      }
      if (_optionalChain([audioBufferCache, 'access', _23 => _23.currentChunk, 'optionalAccess', _24 => _24.next, 'optionalAccess', _25 => _25.next]) === chunk) {
        if (audioBufferCache.pendingBuffer === null) {
          audioBufferCache.pendingBuffer = buffer;
        }
      }
      this.tryResumePlayback();
    }

    // Start chunk decode
     decodeChunk(chunk) {
      if (chunk === null) {
        return;
      }
      if (chunk.ready) {
        chunk.createBuffer().then(
          (buffer) => {
            this.onBufferDecoded(chunk, buffer);
          },
          (err) => {
            this._fire('playbackerror', err);
          }
        );
      } else {
        chunk.once('ready', () => {
          chunk.createBuffer().then(
            (buffer) => {
              this.onBufferDecoded(chunk, buffer);
            },
            (err) => {
              this._fire('playbackerror', err);
            }
          );
        });
      }
    }

    // Attempt to resume playback when not actual Playing
     tryResumePlayback() {
      if (this._actualPlaying || !this.playing) {
        return;
      }
      if (this._audioBufferCache === null) {
        return;
      }
      if (this.audioBufferCacheHit()) {
        this.startPlay();
      }
    }
  }

  class Clone extends Clip {
    

    constructor(original) {
      super({
        context: original.context,
        url: original.url,
        adapter: original.adapter,
      });
      this.original = original;
    }

    buffer() {
      return this.original.buffer();
    }

    clone() {
      return this.original.clone();
    }

    get canplaythrough() {
      return this.original.canplaythrough;
    }

    // eslint-disable-next-line class-methods-use-this
    set canplaythrough(_) {
      // eslint-disable-line no-unused-vars
      // noop
    }

    get loaded() {
      return this.original.loaded;
    }

    // eslint-disable-next-line class-methods-use-this
    set loaded(_) {
      // eslint-disable-line no-unused-vars
      // noop
    }

    get _chunks() {
      return this.original._chunks;
    }

    // eslint-disable-next-line class-methods-use-this
    set _chunks(_) {
      // eslint-disable-line no-unused-vars
      // noop
    }
  }

  function _nullishCoalesce$1(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } }

  // http://mpgedit.org/mpgedit/mpeg_format/mpeghdr.htm
  const bitrateLookup = {
    11: [null, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448],
    12: [null, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384],
    13: [null, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320],
    21: [null, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256],
    22: [null, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160],
  };

  // eslint-disable-next-line prefer-destructuring
  bitrateLookup[23] = bitrateLookup[22];

  function getFrameLength(
    data,
    i,
    metadata
  ) {
    const { mpegVersion } = metadata;
    const { mpegLayer } = metadata;
    const { sampleRate } = metadata;

    const bitrateCode = (data[i + 2] & 0b11110000) >> 4;
    const bitrate =
      (_nullishCoalesce$1(bitrateLookup[`${mpegVersion}${mpegLayer}`][bitrateCode], () => ( 0))) * 1e3;
    const padding = (data[i + 2] & 0b00000010) >> 1;

    const length = ~~(mpegLayer === 1
      ? ((12 * bitrate) / sampleRate + padding) * 4
      : (144 * bitrate) / sampleRate + padding);

    return length;
  }

  // http://www.mp3-tech.org/programmer/frame_header.html
  // frame header starts with 'frame sync' – eleven 1s
  function parseFrameHeader(
    data,
    i = 0
  ) {
    return {
      mpegVersion: data[i + 1] & 0b00001000,
      mpegLayer: data[i + 1] & 0b00000110,
      sampleRate: data[i + 2] & 0b00001100,
      channelMode: data[i + 3] & 0b11000000,
    };
  }

  const mpegVersionLookup = {
    0: 2,
    1: 1,
  };

  const mpegLayerLookup = {
    1: 3,
    2: 2,
    3: 1,
  };

  const sampleRateLookup = {
    0: 44100,
    1: 48000,
    2: 32000,
  };

  const channelModeLookup = {
    0: 'stereo',
    1: 'joint stereo',
    2: 'dual channel',
    3: 'mono',
  };

  function parseMetadata(metadata) {
    const mpegVersion = mpegVersionLookup[metadata.mpegVersion >> 3];

    return {
      mpegVersion,
      mpegLayer: mpegLayerLookup[metadata.mpegLayer >> 1],
      sampleRate: sampleRateLookup[metadata.sampleRate >> 2] / mpegVersion,
      channelMode: channelModeLookup[metadata.channelMode >> 6],
    };
  }

  const validateHeader = (data, i = 0) => {
    // First 11 bits should be set to 1, and the 12nd bit should be 1 if the audio is MPEG Version 1 or
    // MPEG Version 2, this means MPEG Version 2.5 is not supported by this library.
    if (data[i + 0] !== 0b11111111 || (data[i + 1] & 0b11110000) !== 0b11110000)
      return false;

    const valid =
      // Layer Description should not be 00, since it is reserved
      (data[i + 1] & 0b00000110) !== 0b00000000 &&
      // Bitrate should not be 1111, since it is bad
      (data[i + 2] & 0b11110000) !== 0b11110000 &&
      // Sampling Rate Frequency should not be 11, the value is reserved
      (data[i + 2] & 0b00001100) !== 0b00001100 &&
      // Emphasis should not be 10, the value is reserved
      (data[i + 2] & 0b00000011) !== 0b00000010;

    return valid;
  };

  const id3Header = new Uint8Array([0x49, 0x44, 0x33, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]);

  function wrapChunk(
    original
  ) {
    if (original.length >= 3) {
      if (original[0] === 0x49 && original[1] === 0x44 && original[2] === 0x33) {
        // already has ID3 header
        return original
      }
    }
    const result = new Uint8Array(id3Header.length + original.length);
    for (let i = 0; i < id3Header.length; i++) {
      result[i] = id3Header[i];
    }
    for (let i = 0; i < original.length; i++) {
      result[i + id3Header.length] = original[i];
    }
    return result
  }

  function _nullishCoalesce(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } }

  const mp3Adapter = {
    validateChunk: validateHeader,
    getChunkLength: (rawData, metadata, index) => {
      if (metadata === null) {
        return null;
      }
      return getFrameLength(rawData, _nullishCoalesce(index, () => ( 0)), metadata);
    },
    getChunkDuration: (_, metadata) => {
      if (metadata === null) {
        return null;
      }
      return 1152 / metadata.sampleRate;
    },
    getChunkMetadata: (rawData, index) => {
      const rawMetadata = parseFrameHeader(rawData, index);

      if (!rawMetadata) return null;

      return parseMetadata(rawMetadata);
    },
    wrapChunk: (rawData) => {
      return wrapChunk(rawData);
    }
  };

  exports.Clip = Clip;
  exports.mp3Adapter = mp3Adapter;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=browser.js.map
